import json
import os
import unittest
from multiprocessing import Process
from pathlib import Path
from shutil import copyfile

import requests
from urllib3 import Retry

from ..base.accessors import generate_file_accessor

class TestServerBaseClass(unittest.TestCase):
    """
    Base class for unittests of API functionality.  Implements both server and clients for testing.
    """

    app_name = 'model_server.base.api:app'

    def setUp(self) -> None:
        import uvicorn
        host = '127.0.0.1'
        port = 5001

        self.server_process = Process(
            target=uvicorn.run,
            args=(self.app_name, ),
            kwargs={'host': host, 'port': port, 'log_level': 'critical'},
            daemon=True
        )
        self.uri = f'http://{host}:{port}/'
        self.server_process.start()

    def _get_sesh(self):
        sesh = requests.Session()
        retries = Retry(
            total=5,
            backoff_factor=0.1,
        )
        sesh.mount('http://', requests.adapters.HTTPAdapter(max_retries=retries))
        return sesh

    def _get(self, endpoint):
        return self._get_sesh().get(self.uri + endpoint)

    def _put(self, endpoint, query=None, body=None):
        return self._get_sesh().put(
            self.uri + endpoint,
            params=query,
            data=json.dumps(body)
        )

    def tearDown(self) -> None:
        self.server_process.terminate()
        self.server_process.join()

    def copy_input_file_to_server(self):
        resp = self._get('paths')
        pa = resp.json()['inbound_images']
        copyfile(
            self.input_data['path'],
            Path(pa) / self.input_data['name']
        )
        return self.input_data['name']

    def get_accessor(self, accessor_id, filename=None):
        resp = self._put(f'/accessors/write_to_file/{accessor_id}', query={'filename': filename})
        where_out = self._get('paths').json()['outbound_images']
        fp_out = (Path(where_out) / resp.json())
        self.assertTrue(fp_out.exists())
        return generate_file_accessor(fp_out)


def setup_test_data():
    """
    Look for test data, create test output directory, parse and return meta information
    :return:
        meta (dict) of test data and paths
    """
    # places to look for test data
    data_paths = [
        os.environ.get('UNITTEST_DATA_ROOT'),
        Path.home() / 'model_server' / 'testing',
        os.getcwd(),
    ]
    root = None

    # look for first instance of summary.json
    for dp in data_paths:
        if dp is None:
            continue
        sf = (Path(dp) / 'summary.json')
        if sf.exists():
            with open(sf, 'r') as fh:
                meta = json.load(fh)
                root = dp
                break

    if root is None:
        raise Exception('Could not find test data, try setting environmental variable UNITTEST_DATA_ROOT.')

    meta['root'] = Path(root)
    op_ev = os.environ.get('UNITTEST_OUTPUT', (meta['root'] / 'test_output'))
    meta['output_path'] = Path(op_ev)
    meta['output_path'].mkdir(parents=True, exist_ok=True)

    # resolve relative paths
    def _resolve_paths(d):
        keys = list(d.keys())
        for k in keys:
            if k == 'name':
                d['path'] = meta['root'] / d['name']
            elif isinstance(d[k], dict):
                _resolve_paths(d[k])
    _resolve_paths(meta)

    return meta

# object containing test data paths and metadata, for import into unittest modules
meta = setup_test_data()